// deno-lint-ignore-file no-explicit-any ban-types
import { Reflect } from "https://deno.land/x/reflect_metadata@v0.1.12/mod.ts";
import serialize from "./serialize.ts";
import HttpError from "./HttpError.ts";
import RestResponse from "./RestResponse.ts";
import RestRequest from "./RestRequest.ts";

export type HandlerRes = Promise<RestResponse | undefined> | RestResponse | undefined;

export type HttpMethod = '*' | 'GET' | 'POST' | 'CHANGE' | 'DELETE' | 'PUT' | 'UPDATE';

export interface Handler {
    handle(req: RestRequest): HandlerRes;
}
export interface RestOptions {
    route?: string;
}

type ProcessFunc = (req: RestRequest, arg: unknown, name: string) => Promise<unknown> | unknown;
type Base<KeyT extends string> = Router & { [X in KeyT]: Function; };

interface RouterHandler {
    path: string;
    handler: Handler;
}

export function makeParameterModifier(func: ProcessFunc) {
    return (target: Router, key: string, index: number) => {
        let res = Reflect.getOwnMetadata('router:params', target, key);

        if (res === undefined) Reflect.defineMetadata('router:params', res = [], target, key);
        (res[index] ??= []).push(func);

        return res;
    }
}


function addMetaQuery(target: any, ...handlers: ((r: Router) => RouterHandler)[]) {
    let props = Reflect.getOwnMetadata('router:queries', target);
    if (props === undefined) Reflect.defineMetadata('router:queries', props = [], target);

    props.push(...handlers);
}

export function rest<KeyT extends keyof T & string, T extends Base<KeyT>>(method: HttpMethod, route?: string) {
    return (target: T, key: KeyT) => {
        const path = route ?? key;

        addMetaQuery(target, (r: any) => ({
            path, handler: {
                async handle(req) {
                    if (method !== '*' && req.method.toUpperCase() !== method) return undefined;

                    const params: string[] = [];
                    const func = (r as T)[key];
                    const args: any[] = [];
                    const allMods: ([ProcessFunc, ...ProcessFunc[]] | undefined)[] = [];

                    let signature = (target[key] as Function).toString();
                    signature = signature.substring(signature.indexOf('(') + 1, signature.indexOf(')'));
                    params.push(...signature.split(',').map(v => v.trim()).filter(v => v !== ''));

                    for (let proto = target; proto instanceof Router; proto = Object.getPrototypeOf(proto)) {
                        const data = Reflect.getOwnMetadata('router:params', proto, key);
                        if (!data) continue;
                        for (let i = 0; i < data.length; i++) {
                            allMods[i] ??= [] as any;
                            if (data[i]) allMods[i]!.push(...data[i]);
                        }
                    }

                    for (let i = 0; i < params.length; i++) {
                        const param = params[i];

                        let arg: any = req.params[param];

                        for (const mod of allMods[i] ?? []) {
                            arg = await mod(req, arg, params[i]);
                        }

                        args.push(arg);
                    }

                    const res = func.apply(r, args);
                    if (res instanceof RestResponse) return res;
                    return new RestResponse().body(await serialize(res));
                }
            },
        }));
    };
}
export function route<KeyT extends keyof T & string, T extends Router & { [x in KeyT]: Handler }>(path?: string) {
    return (target: T, key: KeyT) => {
        addMetaQuery(target, (r: any) => ({
            handler: r[key],
            path: path ?? key,
        }));
    };
}

export default class Router {
    private _handlers?: RouterHandler[];
    public defaultHandler?: Handler;

    private _init() {
        if (this._handlers === undefined) {
            this._handlers = [];

            // why the actual fuck not
            // deno-lint-ignore no-this-alias
            for (let proto = this; proto instanceof Router; proto = Object.getPrototypeOf(proto)) {
                const props = Reflect.getOwnMetadata('router:queries', proto);
                for (const handler of props ?? []) {
                    this._handlers.push(handler(this));
                }
            }
        }
        return this._handlers;
    }

    public async handle(req: RestRequest) {
        for (const hnd of this._init()) {
            const _req = req.match(hnd.path);
            if (_req) {
                try {
                    const res = await hnd.handler.handle(_req);
                    if (res) return res;
                }
                catch (e) {
                    const res = await this.onError(_req, e);
                    if (res instanceof HttpError) return new RestResponse()
                        .body(await serialize(res.body))
                        .status(res.status);
                    else return res;
                }
            }
        }

        return this.defaultHandler?.handle(req);
    }
    public addHandler(path: string, handler: Handler) {
        this._init().push({ path, handler });
        return this;
    }

    public onError(_req: RestRequest, error: unknown): Promise<RestResponse | HttpError> | RestResponse | HttpError {
        if (error instanceof HttpError) return error;
        else {
            console.error(error);
            try {
                return new HttpError(`Internal error: ${error}\nSee logs for details`, 500);
            }
            catch {
                return new HttpError('Internal error.\nSee logs for details', 500);
            }
        }
    }

    public async attach(server: Deno.Listener) {
        for await (const conn of server) {
            for await (const req of Deno.serveHttp(conn)) {
                const r = await this.handle(await RestRequest.fromMessage(req));
                if (r) req.respondWith(r.toFetchResponse());
            }
        }
    }
}
